# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Sensor list.
"""
from typing import (
    AsyncGenerator
)

from .sensor import G90Sensor
from .base_list import G90BaseList
from ..const import G90Commands


class G90SensorList(G90BaseList[G90Sensor]):
    """
    Sensor list class.
    """
    async def _fetch(self) -> AsyncGenerator[G90Sensor, None]:
        """
        Fetch the list of sensors from the panel.

        :yields: G90Sensor: Sensor entity.
        """
        entities = self._parent.paginated_result(
            G90Commands.GETSENSORLIST
        )

        async for entity in entities:
            yield G90Sensor(
                *entity.data, parent=self._parent, subindex=0,
                proto_idx=entity.proto_idx
            )
